% QuickPsp.m %
% Quicken <phi | Phi> and <beta | Phi> calculations by
% 1. converting the radial function to spline in K-space
% 2. evaluate and tabulate <phi | K> and <beta | K>
% 3. tabulate exp(i * K * R) of all atoms

function [Psp, Atom] = ...
      QuickPsp(Psp, Atom, FFTIndex, number_plane_waves, nGrid, IBZKMax, IBZK, IBZGmin);
  
  fprintf (1, '[QuickPsp]  Tabulating <phi | K> and <beta | K>.\n');
  G = Atom.G;
  TypeMax = Psp{1}.TypeMax;
  atomMax = Atom.AtomMax;
  atom_X  = Atom.X;
  
  ibz=1;
  % GList = zeros(length(WaveFunctionFFTindex(1,1,:)), 3);
  GList = GListGenerator(ibz, FFTIndex, nGrid);
  GList = GList*G';
  Gnorm = sqrt( sum(GList.^2, 2) );
  KMax  = max(Gnorm) * 1.4;
  Kmesh = 64+1;
  Kdelta = KMax / (Kmesh-1);
  Kgrid  =  0 : Kdelta : KMax; 
  
  
  for type = 1 : TypeMax,

    clear rbetaK;
    Gbetamax   = Psp{type}.beta_Gmax;
    Kbetamesh  = Psp{type}.rbetaK_LENGTH;
    Kbetadelta = Gbetamax / (Kbetamesh-1);
    Kbetagrid  = 0 : Kbetadelta : Gbetamax;
    
    Psp{type}.projK  = zeros(Psp{type}.projMax, number_plane_waves, IBZKMax);
    for ibeta = 1 : Psp{type}.betaMax,
      rbetaK = Psp{type}.rbetaK(ibeta,:);
      Psp{type}.rbetaKpp(ibeta) = spline(Kbetagrid, rbetaK);
    end;
    
    
    clear rphiK;
    aux1 = 0;
    ngrid = length(Psp{type}.RadialGrid);
    ngrid = 2*(floor(ngrid/2)+1)-1;
    Psp{type}.RadialGridDiff = Psp{type}.RadialGridAB(1,1:ngrid);
    rgrid = Psp{type}.RadialGrid;
    Psp{type}.orbK  = zeros(Psp{type}.orbMax, number_plane_waves, IBZKMax);
    
    for iphi = 1 : Psp{type}.phiMax,
      rphi = Psp{type}.rphi(iphi,:);
      L = Psp{type}.phi_L(iphi);
      
      igrid = 1;
      if (L==0),
        sphbessel = ones(1,ngrid);
        aux1 = rgrid(1,1:ngrid) .* ...
               rphi(1,1:ngrid) .* sphbessel;
        rphiK(1,igrid) = ...
            RadialIntegralLog(ngrid, aux1, Psp{type}.RadialGridDiff);
      else
        aux1 = rgrid(1,1:ngrid) .* ...
               rphi(1,1:ngrid) .* ...
               SphericalBessel(L,Kgrid(igrid)*rgrid(1,1:ngrid));
        rphiK(1,igrid) = ...
            RadialIntegralLog(ngrid, aux1, Psp{type}.RadialGridDiff);
      end;
      
      for igrid = 2 : Kmesh, % GListLength%
        if (abs(Kgrid(igrid)-Kgrid(igrid-1))< 1e-10),
          rphiK(1,igrid) = rphiK(1,(igrid-1));
        else
          aux1 = rgrid(1,1:ngrid) .* ...
                 rphi(1,1:ngrid) .* ...
                 SphericalBessel(L,Kgrid(igrid)*rgrid(1,1:ngrid));
          rphiK(1,igrid) = ...
              RadialIntegralLog(ngrid, aux1, Psp{type}.RadialGridDiff);
        end;
      end;
      
      Psp{type}.rphiKpp(iphi) = spline(Kgrid, rphiK);
    end;
    
    
    
    
    
    GList = [];
    for ibz = 1 : IBZKMax,
      
      GList = GListGenerator(ibz, FFTIndex, nGrid);
      GList = GList * G';
      
      % possible problem if missing following code %
      % when do <orb | S | orb> integration, or <orb|orb> , %
      % there are some part of GList which is zero -- or not exist    %
      % then for L==0 case, we will get finite value for orbK, that is not right %
      % therefore, later we will set all of these orbK at zero GList range to be 0, except origin point %

      GLength = size(GList,1);
      
      Gmin = IBZGmin(ibz);
      
      KList = GList + ones(number_plane_waves,1) * IBZK(ibz,:);
      
      [theta, phi, Knorm] = ...
          Cartesian2Spherical(KList(:,1), KList(:,2), KList(:,3));
      
      for iorb = 1 : Psp{type}.orbMax,
        iphi = Psp{type}.iphi(iorb);
        L = Psp{type}.orb_L(iorb);
        M = Psp{type}.orb_M(iorb);
        % this mimics HLMPlanewave.m %
        % orbK stands for <orb|k> =<orb(r)|exp(i k r)>  %
        Psp{type}.orbK(iorb, :, ibz) = ...
            ppval(Psp{type}.rphiKpp(iphi), Knorm) .* ...
            conj(HLM(L, M, theta, phi)) * 4 * pi * j^L;
        
      end;
      
      
      for iproj = 1 : Psp{type}.projMax,
        ibeta = Psp{type}.ibeta(iproj);
        L = Psp{type}.proj_L(iproj);
        M = Psp{type}.proj_M(iproj);
        % this mimics HLMPlanewave.m %
        % projK stands for <proj | k>  %
        Psp{type}.projK(iproj, :, ibz) = ...
            ppval(Psp{type}.rbetaKpp(ibeta), Knorm) .* ...
            conj(HLM(L, M, theta, phi)) *j^L/sqrt(Atom.Volume);
%         [rbetakout] = splineproj(Psp{type}.rbetaK(Psp{type}.ibeta, :), Psp{type}.beta_Gmax, Atom.Volume, Knorm); %
%         Psp{type}.projK(iproj, :, ibz) = rbetakout .* conj(HLM(L, M, theta, phi)') * j^L; %
      end;
      
      for iatom = 1 : atomMax,
        atom_phase(iatom, :, ibz) = exp(KList * atom_X(iatom,:)' * j);
      end;
      
      
      if (Gmin<GLength),
        
        Psp{type}.orbK(:,(Gmin+1):GLength,ibz)  = 0.;
        Psp{type}.projK(:,(Gmin+1):GLength,ibz) = 0.;
        atom_phase(:, (Gmin+1):GLength, ibz) = 0.;
        
      end;
      
      
    end;
    
    
    
  end;
  
  clear rphiK ;
  
  Atom.Phase = atom_phase;
  
  fprintf (1, '\n');
  
  
  
