% QuickWavefunction.m %

function [PW, Bloch, Psp] = QuickWavefunction(PW, Bloch, G, H, Psp); 

  fprintf (1, '\n[QuickWavefunction]  Calculating FFT Index according to lattice point group\n');
  
  WaveFunctionFFTindexIBZ = PW.FFTIndex;
  Enk        = PW.EigenValues;
  LatRot     = PW.LatRot;
  CartRot    = PW.CartRot;
  IBZK       = PW.IBZK;
  IBZKWeight = PW.IBZKWeight;
  IBZKMax    = PW.IBZKMax;

  number_of_spin    = PW.nSpins;
  number_of_bands   = PW.nBands;
  OccupationNumbers = PW.Occupation;
  BZKMax            = PW.BZKMax;
  BZKIN             = PW.BZK;
  MultiplierGmin    = PW.IBZGmin;
  BZK               = PW.BZK;
  oldPsp            = Psp;
  
  WaveFunctionFFTindex = [];

  NG0 = max(squeeze(WaveFunctionFFTindexIBZ(1,:,:))');
  NG  = NG0;
  if (IBZKMax>=2)
    for ibz = 2: IBZKMax
      NG = max(squeeze(WaveFunctionFFTindexIBZ(ibz,:,:))');
      if (NG0 ~= NG)
        fprintf(1,'\n[QuickWavefunction]  G-grid in WaveFunctionFFTindex has problem.\n');
        pause;
      end;
    end;
  end;
  

  Multiplier = WaveFunctionFFTindexIBZ-1;
  for dim = 1 : 3,
    Multiplier(:,dim,:) = Multiplier(:,dim,:) - NG(dim) * ...
        ( Multiplier(:,dim,:) > (NG(dim)/2.) );
  end;
  

  
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
  newIBZKMax = 0;
  newIBZK    = [];
  newIBZKWeight     = [];
  newMultiplierGmin =[];

  for ibz = 1 : IBZKMax,
    
    [bzk, latrotation] = IBZ2BZ(ibz, IBZK,BZK, IBZKWeight, LatRot, CartRot, G, H);
    lenbz = size(bzk,1);
    
    newIBZK = [newIBZK; bzk];
    
    totallen = size(Multiplier(ibz,:,:), 3 );
    
    for nbz = 1 : lenbz,
      tmpmul = squeeze(latrotation(:,:,nbz)) * ...
               squeeze(Multiplier(ibz,1:3,:));
      
      if (norm(IBZK(ibz,:)) > 1e-6),
        newIBZKWeight(newIBZKMax+nbz) = 2;
      else
        newIBZKWeight(newIBZKMax+nbz) = 1;
      end;
      
      for dim = 1 : 3,
        WaveFunctionFFTindex(newIBZKMax+nbz, dim, 1:totallen) = ...
            tmpmul(dim,:) + NG(dim) * ( tmpmul(dim,:) < 0 ) + 1;      
      end;
      
      newMultiplierGmin(newIBZKMax+nbz) = MultiplierGmin(ibz);
      
    end;
    
    if (MultiplierGmin(ibz) < totallen),
      WaveFunctionFFTindex((newIBZKMax+1) : (newIBZKMax+nbz), :, ...
                           (MultiplierGmin(ibz)+1):totallen) =0;
    end;
    
    newIBZKMax = newIBZKMax + lenbz;
    
  end;
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  
  
  
  newBloch = cell(newIBZKMax, number_of_spin, number_of_bands);
  newEnk   = zeros(newIBZKMax, number_of_spin, number_of_bands);
  newOccupationNumbers = zeros(newIBZKMax, number_of_spin, number_of_bands);
  
  % add 05-15-2006 %
  newLatrot =  zeros(3,3, newIBZKMax);
  
  newIBZKMax = 0;
  in = 1;
  for ibz = 1 : IBZKMax,
    
    [bzk, latrotation] = IBZ2BZ(ibz, IBZK,BZK, IBZKWeight, LatRot, CartRot, G, H);
    lenbz = size(bzk,1);
    
    
    % add 05-15-2006 %
    newLatrot(:,:,in:(in+lenbz-1)) = latrotation;
    in = in + lenbz;
    
    for nbz = 1 : lenbz,
      for ispin = 1 : number_of_spin,
        for iband = 1 : number_of_bands,
          newBloch{newIBZKMax+nbz,ispin,iband}.a = Bloch{ibz,ispin,iband}.a;
          newOccupationNumbers(newIBZKMax+nbz,ispin,iband) = ...
              OccupationNumbers(ibz,ispin, iband);  
          newEnk(newIBZKMax+nbz,ispin,iband) = Enk(ibz,ispin, iband);
        end;
      end;
    end;
    
    newIBZKMax = newIBZKMax + lenbz;
  end;

  
  newIBZKWeight = newIBZKWeight / sum(newIBZKWeight);
  
  if ( norm(BZKMax - newIBZKMax*2) > (1.0 + 1e-4) ),
    fprintf(1, '\n');
    fprintf(1, 'QuickWavefunction: Error time inversion symmetry dimension\n');
    pause;
  end;
  
  
  fprintf (1, '\n');
  
  oldIBZKMax = IBZKMax;
  oldIBZK    = IBZK;
  oldIBZKWeight = IBZKWeight;
  IBZKMax    = newIBZKMax;
  IBZK       = newIBZK;
  IBZKWeight = newIBZKWeight;
  
  oldBloch = Bloch;
  
  Bloch    = [];
  Bloch    = cell(newIBZKMax, number_of_spin, number_of_bands);
  Bloch    = newBloch;
  
  
  PW.Occupation = [];
  PW.IBZK       = [];
  PW.IBZKMax    = [];
  PW.IBZKWeight = [];
  PW.IBZGmin    = [];
  PW.FFTIndex   = [];
  
  PW.EigenValues= newEnk;
  PW.Occupation = newOccupationNumbers;
  PW.IBZK       = newIBZK;
  PW.IBZKMax    = newIBZKMax;
  PW.IBZKWeight = newIBZKWeight;
  PW.IBZGmin    = newMultiplierGmin;
  PW.nIBZKs     = newIBZKMax;
  PW.FFTIndex   = WaveFunctionFFTindex;
  
  PW.oldIBZK    = oldIBZK;
  PW.oldIBZKMax = oldIBZKMax;
  PW.oldIBZKWeight = oldIBZKWeight;
  

  
  oldEnk = [];
  oldBloch = [];
  
  newEnk = [];
  newBloch = [];
  newOccupationNumbers = [];
  newIBZKMax = [];
  
  
  return;
  
