% HLM.m : unitary transformed spherical harmonics 
% that are easier for analysis in Cartesian frame.
%  NOTE :   Y_LM (G=0) = SQRT(FPI)  WHEN L=0 %
%                      = 0          WHEN L>0 %

% form 1 %
% HLM(L,M) = \frac{1}{\sqrt{2}} ( Y_{l, -m} + (-1)^m Y^{*}_{l,m} ) %
%          = \frac{i}{\sqrt{2}} ( Y_{l, -m} - (-1)^m Y^{*}_{l,m} ) %

% form 2 %
% HLM(L,M) = \frac{ (-1)^m }{ \sqrt{2}} ( Y_{l, m} + Y^{*}_{l,m} )  (m>0) %
%          = \frac{-i (I-1)^m }{\sqrt{2}} ( Y_{l, -m} - Y^{*}_{l,-m} )  (m<0) %
%          = Y_{l,m}           (m=0) %

% http://www1.elsevier.com/homepage/saa/eccc3/paper48/eccc3.html %

function value = HLM(L, M, theta, phi)
  
  x = sin(theta) .* cos(phi);
  y = sin(theta) .* sin(phi);
  z = cos(theta);
  fpi = 4*pi;
  % x,y,z are normalized by r   %

  if L == 0,
    value = 1/sqrt(4*pi) .* ones(length(theta),1);
    
  elseif L == 1,
    if M == -1,
      value = sqrt(0.75 / pi) * x;              % X
    elseif M == 0,
      value = sqrt(0.75 / pi) * y;              % Y
    elseif M==1,
      value = sqrt(0.75 / pi) * z;              % Z
    else
      fprintf(1,'M in [-L, .. ,+L].   Wrong (L,M) = (%d,%d)',L,M);
    end;
    
  elseif L == 2,
    if M == -2,
      c = sqrt(15/fpi);                         
      value = c*x.*y;                           % X*Y
    elseif M==-1,
      c = sqrt(15/fpi);
      value = c*x.*z;                           % X*Z
    elseif M==0,
      c = sqrt(5/4/fpi);
      value = c*(3*(z.^2) - 1);                 % (3*Z*Z -1)
    elseif M==1,
      c = sqrt(15/fpi);
      value = c*y.*z;                           % Y*Z
    elseif M==2,
      c = sqrt(15/4/fpi);
      value = c*(x.^2 - y.^2);                  % X*X - Y*Y
    else
       fprintf(1,'M in [-L, .. ,+L].   Wrong (L,M) = (%d,%d)',L,M);
    end;
    
  elseif L==3,
    if M == -3,
      c = sqrt(7/fpi)*5/2;
      value = c*(x.^3 - 0.6*x);                 % X^3 - 3/5*X
    elseif M==-2,
      c = sqrt(7/fpi)*5/2;
      value = c*(y.^3 - 0.6*y);                 % Y^3 - 3/5*Y
    elseif M==-1,
      c = sqrt(7*15/fpi);
      value = c*(x.*y.*z);                      % X*Y*Z
    elseif M==0,
      c = sqrt(7/fpi)*5/2;
      value = c*(z.^3 - 0.6*z);                 % Z^3 - 3/5*Z
    elseif M==1,
      c = sqrt(7*15/fpi)/2;
      value = c*z.*(x.^2-y.^2);                 % Z*(X*X - Y*Y)
    elseif M==2,
      c = sqrt(7*15/fpi)/2;
      value = c*y.*(z.^2-x.^2);                 % Y*(Z*Z - X*X)
    elseif M==3,
      c = sqrt(7*15/fpi)/2;
      value = c*x.*(y.^2-z.^2);                 % X*(Y*Y - Z*Z)
    else
      fprintf(1,'M in [-L, .. ,+L].   Wrong (L,M) = (%d,%d)',L,M);
    end;
    
  elseif L==4,
    if M==-4,
      c = sqrt(3*7/fpi)*5/4;
      value = c*(x.^4 + y.^4 + z.^4 - 0.6);     % X^4 + Y^4 + Z^4 - 0.6
    elseif M==-3,
      c = sqrt(9*35/fpi)/2;
      value = c*y.*z.*(y.^2 - z.^2);            % Y*Z*(Y^2 - Z^2)
    elseif M==-2,
      c = sqrt(9*35/fpi)/2;
      value = c*x.*z.*(z.^2 - x.^2);            % Z*X*(Z^2 - X^2)
    elseif M==-1,
       c = sqrt(9*5/fpi)/4;
       value = c*((x.^4-y.^4) - 6* (z.^2).*(x.^2-y.^2));    % (X^4 - Y^4) - 6*Z*Z* (X^2 - Y^2)
    elseif M==0,
      c = sqrt(9*35/fpi)/2;
      value = c*x.*y.*(x.^2-y.^2);              % X*Y*(X^2 - Y^2)
    elseif M==1,
      c = sqrt(9*5/fpi)*7/2; 
      value = c*x.*y.*(z.^2 - 1/7);             % X*Y*(Z^2 - 1/7)
    elseif M==2,
      c = sqrt(9*5/fpi)*7/2;
      value = c*x.*z.*(y.^2 - 1/7);             % X*Z*(Y^2 - 1/7)
    elseif M==3,
      c = sqrt(9*5/fpi)*7/2;
      value = c*y.*z.*(x.^2 - 1/7);             % Y*Z*(X^2 - 1/7)
    elseif M==4,
      c = sqrt(9*5/fpi/3)*7/2;
      value = c*(z.^4 - ...
                 0.5*(x.^4 + y.^4) - ...
                 6/7*(z.^2 - 0.5*(x.^2+y.^2))); % Z^4 - 0.5*(X^4 + Y^4) - 6/7*(Z*Z - 0.5*(X*X + Y*Y))
    else
      fprintf(1,'M in [-L, .. ,+L].   Wrong (L,M) = (%d,%d)',L,M);
    end;
    
  else  %L>=5, not supported yet
    fprintf(1,'HLM:  L in [0,1,2,3,4] . L exceeds this region');
    value = 0;
  end;
  
